/*
Copyright (C) 2010 Udo Klein

This program is free software;
you can redistribute it and/or modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 3 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses>.
*/



//#define DEBUG 1

// 2010 07 11 measured influence of timer / prescaler settings on power consumption
// 2010 07 11 put prescaler definitions outside of the code
// 2010 07 11 switched from int0 level interrupt to pin change interrupt in order to allow for any pin as input
// 2010 07 10 pushbutton code
// 2010 07 10 timer setup
// 2010 07 10 state engine 
// 2010 07 10 build environment


// PDIP
//                                      *--U--+
// (PCINT5/RESET/ADC0/dW)        PB5   -|     |-  VCC
// (PCINT3/XTAL1/CLKI/OC1B/ADC3) PB3   -|     |-  PB2 (SCK/USCK/SCL/ADC1/T0/INT0/PCINT2) 
// (PCINT4/XTAL2/CLKO/OC1B/ADC2) PB4   -|     |-  PB1 (MISO/DO/AIN1/OC0B/OC1A/PCINT1)
//                               GND   -|     |-  PB0 (MOSI/DI/SDA/AIN0/OC0A/OC1A/AREF/PCINT0)
//                                      +-----+


enum delay_constants {	switch_on    =  7,  // how many 1/10s to push before switching on
						no_switch_on = 20,  // how many 1/10s to hold after switching on before falling back to off
						switch_off   =  7   // how many 1/10s to push before switching off
					 };
							

#define BUTTON_DDR  DDRB
#define BUTTON_PORT PORTB
#define BUTTON_PIN  PINB
#define BUTTON_INPUT_BIT  4
#define BUTTON_ISR_VECT PCINT0_vect


#define OUTPUT0_DDR  DDRB
#define OUTPUT0_PORT PORTB
#define OUTPUT0_BIT  3
// Attention: CKDIV Fuse must be set to 1
//            Clock Fuse must be set 128kHz WDT

#include <avr/power.h>

enum t0_prescaler_constants {	t0_no_clock          = (0 << CS02) | (0 << CS01) | (0 << CS00),
								t0_prescale_by_1     = (0 << CS02) | (0 << CS01) | (1 << CS00),
								t0_prescale_by_8     = (0 << CS02) | (1 << CS01) | (0 << CS00),
								t0_prescale_by_64    = (0 << CS02) | (1 << CS01) | (1 << CS00),
								t0_prescale_by_256   = (1 << CS02) | (0 << CS01) | (0 << CS00),
								t0_prescale_by_1024  = (1 << CS02) | (0 << CS01) | (1 << CS00),
								t0_ext_clock_falling = (1 << CS02) | (1 << CS01) | (0 << CS00),
								t0_ext_clock_rising  = (1 << CS02) | (1 << CS01) | (1 << CS00)};
								
enum system_prescaler_constants {	system_prescale_1   = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0),
									system_prescale_2   = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (1 << CLKPS0),
									system_prescale_4   = (0 << CLKPS3) | (0 << CLKPS2) | (1 << CLKPS1) | (0 << CLKPS0),
									system_prescale_8   = (0 << CLKPS3) | (0 << CLKPS2) | (1 << CLKPS1) | (1 << CLKPS0),
									system_prescale_16  = (0 << CLKPS3) | (1 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0),
									system_prescale_32  = (0 << CLKPS3) | (1 << CLKPS2) | (0 << CLKPS1) | (1 << CLKPS0),
									system_prescale_64  = (0 << CLKPS3) | (1 << CLKPS2) | (1 << CLKPS1) | (0 << CLKPS0),
									system_prescale_128 = (0 << CLKPS3) | (1 << CLKPS2) | (1 << CLKPS1) | (1 << CLKPS0),
									system_prescale_256 = (1 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0)};

// 204uA pushed / 0.4uA not pushed 128kHz WDT
// using a smaller prescaler for t0 will result in more power consumption
// using a smaller prescaler for system clock would decrease power consumption further but it will make programming a major pain
enum prescalers { system_prescaler = system_prescale_1,   // forces to flash @6.5kHz or lower
			     t0_prescaler     = t0_prescale_by_1024  }; // 125 Hz



// prescaling further will decrease the current consumption only maginally
// however for some unknow reason this will confuse my ISP --> I will not push this further

// in order to consume even less power it would be more reasonable to switch to the 128kHz oscillator
// this should end up around 100-150uA while pushed

// since pushed time is always only some seconds this will not result in significant gains though
// I stick to this setup


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
//#include <avr/wdt.h>
#include <stdlib.h>
#include <avr/wdt.h>

#define read_bit(value, bit) (((value) >> (bit)) & 1)
#define set_bit(value, bit) ((value) |= (1UL << (bit)))
#define clear_bit(value, bit) ((value) &= ~(1UL << (bit)))
#define toggle_bit(value, bit) ((value) ^= ~(1UL << (bit)))
#define write_bit(value, bit, bit_value) (bit_value ? set_bit(value, bit) : clear_bit(value, bit))


enum io_constants        {low    = 0, high   = 1};
enum ddr_constants       {input  = 0, output = 1};
enum button_constants    {pushed = 0, open   = 1};


// only to wakeup
EMPTY_INTERRUPT(TIMER0_COMPA_vect);
EMPTY_INTERRUPT(BUTTON_ISR_VECT);

// memory consumption is no issue but performace is an issue
// since performance is directly tied to power consumption
// --> by default everything shall be inlined

inline void nop() {	asm volatile ("nop"::); }


inline void timer0_init() {
	// power on timer 0
	clear_bit(PRR, PRTIM0);

	// set timer 0 CTC mode
	// disable all other stuff
	TCCR0A = 1 << WGM01;

	// power off timer 0
	set_bit(PRR, PRTIM0);
}

void sleep_1s_div_10() __attribute__((noinline));
void sleep_1s_div_10() {
	// power on timer 0
	clear_bit(PRR, PRTIM0);

	TCCR0B = t0_prescaler;        // 125 Hz
	OCR0A = 11;                   // 12 = 11+1 Ticks --> ~0.1s

	// reset timer 0 prescaler
	set_bit(GTCCR, PSR0);

	// reset counter
	TCNT0 = 0;

	// OCIE0A: Timer/Counter0 Output Compare Match A Interrupt Enable
	TIMSK = 1 << OCIE0A;

	// goto sleep till ls timer interrupt triggered
	set_sleep_mode(SLEEP_MODE_IDLE);
	sleep_enable();
	sei();
	sleep_cpu();
	sleep_disable();
	cli();

	// stop timer 0
	TCCR0B = t0_no_clock;

	// disable timer0 interrupts
	TIMSK = 0x00;       // disable timer interrupts

	// power off timer 0
	set_bit(PRR, PRTIM0);
}


// button handling is trickier than it looks
//   - the input must have a defined value to avoid undesired switching
//   - we could pull it by default but then current consumption for pushing the button may be as high as 550uA
//   - so we have to not pull it until we want to evaluate it or until we want to prepare it for interrupts
//     - the input buffers will require 1.5 cycles till they see what happened
//     - we need to compensate for possible gate charge times
//     --> after enabling the pullup we have to wait a little bit
inline void button_disable() {
	// set button to low output --> will not consume any power if button is connected to "low"
	set_bit(BUTTON_PORT, low);
	set_bit(BUTTON_DDR, output);
}

inline void button_enable() {
	// Set button to input and enable pullup --> need to wait ~4 cycles before anything
	// reasonable can be said. In theory it might be ready earlier, but gate charge
	// times must be considered --> we stay clear if we take more than 4 cycles.
	set_bit(BUTTON_DDR, input);
	set_bit(BUTTON_PORT, high);
}


inline void button_init() {
	// disable button interrupts
	GIMSK = 0x00;
	
	button_disable();	
}

uint8_t button_state() {

	button_enable();
	// wait for input pin to get to high state (if not pushed)
	nop();	nop();	nop();	nop();  nop();
	uint8_t button_pin_state = BUTTON_PIN;
	// disable button in order to avoid current drain
	button_disable();

	return (button_pin_state >> BUTTON_INPUT_BIT ) & 1;  // normalize output to 0/1, 0 = pushed
}


void sleep_till_button_pushed() {

	// we enable early to ensure that gates will be properly precharded
	button_enable();
	
	cli();


	// goto sleep till INTO interrupt triggered
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);

	// clear all pending interrupt flags
	GIFR  = 0xff;	
	// enable pin change interuppts
	GIMSK = 1<<PCIE;

	// set pin change interrupt mask, right now we have 4 cycles delay since button enabled --> enough
	PCMSK = 1<<BUTTON_INPUT_BIT;
	
	if (read_bit(BUTTON_PIN, BUTTON_INPUT_BIT)) {
		// button not pushed --> we can sleep till it becomes pushed
		// otherwise we must not sleep at all, because we will not detect the desired edge
		
		// see comments to avr-libc and the datasheet on why this works
		// there are two important points
		// - the statement after sei() will be always executed
		// - the PCINT will be queued until triggered
		sleep_enable();
		sei();
		sleep_cpu();
		sleep_disable();
		cli();
	}
	
	button_disable();

	// disable button interrupts
	GIMSK = 0x00;
	PCMSK = 0x00;
}

inline void output0_disable() {
	write_bit(OUTPUT0_DDR,  OUTPUT0_BIT, output);
	write_bit(OUTPUT0_PORT, OUTPUT0_BIT, high);
}

inline void output0_enable() {
	write_bit(OUTPUT0_DDR,  OUTPUT0_BIT, output);
	write_bit(OUTPUT0_PORT, OUTPUT0_BIT, low);
}


#ifdef DEBUG
void blink(uint8_t times) {

	for (uint8_t i = 0; i < times; ++i) {
		sleep_1s_div_10();
		output0_enable();
		sleep_1s_div_10();
		output0_disable();
	}
}
#endif


inline void init() {
	// except while sleeping interrupts shall be disabled
	cli();

	// divide main clock by 8
	CLKPR = 1 << CLKPCE;	
	CLKPR = system_prescaler;	

	// ensure there will be no dangling pins
	// set all ports to input
	DDRB = input *0xff;
	// pull all pins high
	PORTB = high *0xff;

	button_init();
	
	output0_disable();

	timer0_init();

	// shut down all peripherals 
	PRR = 0xff;

}


int main()  {

	init();

// basically we now run a state engine with 6 states 0-5
//
// 0 (sleep, output off)                  "full shutdown" 
//   - push button up->down ==> goto 1
// 1 (idle, output off)                   "going to activate"
//   - push button down->up ==> goto 0
//   - longer than 1s idle  ==> goto 2
// 2 (idle, output on)                    "preliminary active"
//   - longer than 2s idle  ==> goto 5
//   - push button down->up  ==> goto 3
// 3 (sleep, output on)                   "fully active"
//   - push button up->down ==> goto 4
// 4 (idle, output on)                    "going to deactivate"
//   - push button down->up ==> goto 3
//   - longer than 1s idle  ==> goto 5
// 5 (idle, output off)                   "preliminary inactive"
//   - push button down->up ==> goto 0

// - we will never power down during pushed button periods
// - this is in order to poll the button_constants
// - polling consumes less current than powering down
// - reason is that powering down would require to activate
//   the pullup thus leaking current through the pullup all the time
	
	#ifdef DEBUG
	blink(3);
	#endif

	// 0 (sleep, output off)                  "full shutdown"
	//   - push button up->down ==> goto 1
	state_0:
		output0_disable();

		sleep_till_button_pushed();
		goto state_1;

	// 1 (idle, output off)                   "going to activate"
	//   - push button down->up ==> goto 0
	//   - longer than 1s idle  ==> goto 2
	state_1:
		output0_disable();

		for (uint8_t i=0; i < switch_on; ++i) {
			sleep_1s_div_10();
			if (button_state() == high) {
				goto state_0;
			}
		}
		goto state_2;

	// 2 (idle, output on)                    "preliminary active"
	//   - longer than 2s idle  ==> goto 5
	//   - push button down->up  ==> goto 3
	state_2:
		output0_enable();

		for (uint8_t i=0; i < no_switch_on; ++i) {
			sleep_1s_div_10();
			if (button_state() == high) {
				goto state_3;
			}
		}
		goto state_5;

	// 3 (sleep, output on)                   "fully active"
	//   - push button up->down ==> goto 4
	state_3:
		output0_enable();

		sleep_till_button_pushed();
		goto state_4;


	// 4 (idle, output on)                    "going to deactivate"
	//   - push button down->up ==> goto 3
	//   - longer than 1s idle  ==> goto 5
	state_4:
		output0_enable();

		for (uint8_t i=0; i < switch_off; ++i) {
			sleep_1s_div_10();
			if (button_state() == high) {
				goto state_3;
			}
		}
		goto state_5;

	// 5 (idle, output off)                  "preliminary inactive"
	//   - push button down->up ==> goto 0
	state_5:
		output0_disable();

		while (true) {
			sleep_1s_div_10();
			if (button_state() == high) {
				goto state_0;
			}
		}
}
